#!build/env/bin/python
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.

import os
import logging
import socket
import psutil
import re
import subprocess
try:
  import json
except:
  import simplejson as json

from cmf.syssql.table import SimpleModule

__VERSION__ = '0.1'

LOG = logging.getLogger(__name__)

hostname = socket.getfqdn()

if os.environ.get("JAVA_HOME") != None:
  JPS = os.path.join(os.environ.get("JAVA_HOME"), "bin", "jps")
  JAVA = os.path.join(os.environ.get("JAVA_HOME"), "bin", "java")
  JSTACK = os.path.join(os.environ.get("JAVA_HOME"), "bin", "jstack")
else:
  JPS = "jps"
  JAVA = "java"
  JSTACK = "jstack"

JTOOL_JAR = os.path.join(os.path.dirname(__file__), "..", "..", "jtool-0.1-SNAPSHOT.jar")

def jps():
  sub = subprocess.Popen([JPS, "-l"], stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
  stdout, stderr = sub.communicate()
  if sub.returncode != 0:
    LOG.info("jps failed: %d", sub.returncode)
    return
  for row in stdout.splitlines():
    pid, name = row.split(" ", 2)
    yield hostname, int(pid), name

def prefix(pid):
  # jtool only works when running as the same user as the pid
  try:
    process = psutil.Process(pid)
    if process.uids.real == os.getuid():
      return []
    else:
      if os.getuid() == 0:
        return [ "sudo", "-u", process.username ]
  except:
    LOG.warning("Failed to find process java pid: " + pid)
    return None


def jmx(pid):
  # jtool only works when running as the same user as the pid
  pre = prefix(pid)
  if pre is None:
    return

  sub = subprocess.Popen(pre + [JAVA, "-jar", JTOOL_JAR, "jmx", str(pid)],
    stdin=None,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    close_fds=True)
  stdout, stderr = sub.communicate()
  if sub.returncode != 0:
    LOG.info("jtool jmx failed: %d %s", sub.returncode, stderr)
    return
  start = stdout.find("{")
  if start == -1:
    LOG.info("jtool jmx failed: %s", start)
    return
  j = stdout[start:]
  data = json.loads(j)
  for bean in data.get("beans"):
    name = bean.pop("name")
    for k, v in flatten_json(bean):
      yield name, k, v

THREAD_RE = re.compile(r'^"?(?P<name>.*)" (?P<daemon>daemon )?prio=(?P<foo>\d+) tid=[0-9a-fx]* nid=(?P<nid>[0-9a-fx]*) (?P<state>.*) (?:\[[0-9a-fx]*\])$(?:\n.*java.lang.Thread.State.*)?(?:\n.*at (?P<at>.*)$)?', re.MULTILINE)

def jstack(pid):
  """
  Parses jstack output, which clearly sucks.

  You can't easily do much better, because the JVM exposes
  basically the string of the jstack output, without much
  typed information.  You can get at a lot of stuff via
  a management agent, but you can't get the 'nid' (the native
  thread id), which is vastly annoying.  (You can, if you use
  "sa-jdi," but that API depends on ptrace, which is increasingly
  disabled.)

  TODO: perhaps ptrace isn't disabled if you're root, and
  that's a shenagan that could be tried?
  """
  pre = prefix(pid)
  if pre is None:
    return

  sub = subprocess.Popen(pre + [JSTACK, str(pid)],
    stdin=None,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    close_fds=True)
  stdout, stderr = sub.communicate()
  if sub.returncode != 0:
    LOG.info("jstack failed: %d %s", sub.returncode, stderr)
    return
  for name, daemon, prio, nid, state, at in THREAD_RE.findall(stdout):
    nid = int(nid, 16)
    yield name, daemon, prio, nid, state, at

class JvmsModule(SimpleModule):
  name = "jvms"

  columns = [
    "hostname",
    "pid",
    "name"
  ]

  def generate(self):
    return iter(jps())

class JmxModule(SimpleModule):
  name = "jmx"

  columns = [
    "hostname",
    "pid",
    "name",
    "key",
    "value"
  ]

  def generate(self):
    for _, pid, name in jps():
      if name == "sun.tools.jps.Jps":
        continue
      for n, k, v in jmx(pid):
        yield hostname, pid, n, k, v

class JstackModule(SimpleModule):
  name = "jstacks"

  columns = [
    "hostname",
    "pid",
    "nid",
    "name",
    "frame",
    "state"
  ]

  def generate(self):
    for _, pid, name in jps():
      if name == "sun.tools.jps.Jps":
        continue
      for name, daemon, prio, nid, state, at in jstack(pid):
        yield hostname, pid, nid, name, at, state


def flatten_json(datum):
  """Maps a nested structure into a flat one."""
  stack = []
  target = []
  flatten_json_helper(datum, stack, target)
  return target

def flatten_json_helper(datum, stack=None, target=None):
  if isinstance(datum, (list, tuple,)):
    for i, item in enumerate(datum):
      stack.append(str(i))
      flatten_json_helper(item, stack, target)
      stack.pop()
  elif isinstance(datum, dict):
    for k, v in datum.iteritems():
      stack.append(k)
      flatten_json_helper(v, stack, target)
      stack.pop()
  else:
    target.append([":".join(stack), str(datum)])
